from abc import ABC, abstractmethod
from enum import Enum
from typing import List, Optional, Callable

import numpy as np

from sampo.scheduler.resource.base import ResourceOptimizer
from sampo.scheduler.resource.coordinate_descent import CoordinateDescentResourceOptimizer
from sampo.scheduler.timeline.base import Timeline
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph
from sampo.schemas.resources import Worker
from sampo.schemas.schedule import Schedule
from sampo.schemas.schedule_spec import ScheduleSpec, WorkSpec
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.schemas.works import WorkUnit
from sampo.utilities.base_opt import dichotomy_int


class SchedulerType(Enum):
    Topological = 'topological'
    HEFTAddEnd = 'heft_add_end'
    HEFTAddBetween = 'heft_add_between'
    Genetic = 'genetic'


class Scheduler(ABC):
    scheduler_type: SchedulerType
    resource_optimizer: ResourceOptimizer

    def __init__(self,
                 scheduler_type: SchedulerType,
                 resource_optimizer: ResourceOptimizer = CoordinateDescentResourceOptimizer(dichotomy_int),
                 work_estimator: Optional[WorkTimeEstimator] = None):
        self.scheduler_type = scheduler_type
        self.resource_optimizer = resource_optimizer
        self.work_estimator = work_estimator

    def schedule(self, wg: WorkGraph,
                 contractors: List[Contractor],
                 spec: ScheduleSpec = ScheduleSpec(),
                 validate: bool = False,
                 timeline: Timeline | None = None) \
            -> Schedule:
        return self.schedule_with_cache(wg, contractors, spec, validate, timeline)[0]

    @abstractmethod
    def schedule_with_cache(self, wg: WorkGraph,
                            contractors: List[Contractor],
                            spec: ScheduleSpec = ScheduleSpec(),
                            validate: bool = False,
                            timeline: Timeline | None = None) \
            -> tuple[Schedule, Time, Timeline]:
        ...

    @staticmethod
    def optimize_resources_using_spec(work_unit: WorkUnit, worker_team: List[Worker], work_spec: WorkSpec,
                                      optimize_lambda: Callable[[np.ndarray], None] = lambda _: None):
        """
        Applies worker team spec to optimization process.
        Can use arbitrary heuristics to increase spec handling efficiency.
        :param work_unit: current work unit
        :param worker_team: current worker team from chosen contractor
        :param work_spec: spec for given work unit
        :param optimize_lambda: optimization func that should hold optimization
            data in its closure and run optimization process when receives `optimize_array`.
            Passing None or default value means this function should only apply spec.
        """
        if len(work_spec.assigned_workers) == len(work_unit.worker_reqs):
            # all resources passed in spec, skipping optimize_resources step
            for w in worker_team:
                w.count = work_spec.assigned_workers[w.name]
        else:
            # create optimize array to save optimizing time
            # this array should contain True if position should be optimized or False if shouldn't
            optimize_array = None
            if work_spec.assigned_workers:
                optimize_array = []
                for w in worker_team:
                    spec_count = work_spec.assigned_workers.get(w.name, 0)
                    if spec_count > 0:
                        w.count = spec_count
                        optimize_array.append(False)
                    else:
                        optimize_array.append(True)

                optimize_array = np.array(optimize_array)

            optimize_lambda(optimize_array)



from typing import Dict, List, Tuple, Optional

from sampo.scheduler.heft.time_computaion import calculate_working_time
from sampo.scheduler.timeline.base import Timeline
from sampo.schemas.contractor import WorkerContractorPool, Contractor
from sampo.schemas.graph import GraphNode
from sampo.schemas.resources import Worker
from sampo.schemas.scheduled_work import ScheduledWork
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.schemas.types import AgentId


class JustInTimeTimeline(Timeline):

    def __init__(self, worker_pool: WorkerContractorPool):
        self._timeline = {}
        # stacks of time(Time) and count[int]
        for worker_type, worker_offers in worker_pool.items():
            for worker_offer in worker_offers.values():
                self._timeline[worker_offer.get_agent_id()] = [(Time(0), worker_offer.count)]

    def find_min_start_time(self, node: GraphNode, worker_team: List[Worker],
                            id2swork: Dict[GraphNode, ScheduledWork],
                            work_estimator: Optional[WorkTimeEstimator] = None) -> Time:
        """
        Define the nearest possible start time for current job. It is equal the max value from:
        1. end time of all parent tasks
        2. time previous job off all needed workers to complete current task

        :param node: target node
        :param worker_team: worker team under testing
        :param id2swork:
        :param work_estimator:
        :return: found time, queue of employees without assigned workers
        """
        # if current job is the first
        if len(id2swork) == 0:
            return Time(0)
        # define the max end time of all parent tasks
        max_parent_time = max([id2swork[parent_node].finish_time
                               for parent_node in node.parents], default=Time(0))
        # define the max agents time when all needed workers are off from previous tasks
        max_agent_time = Time(0)

        # For each resource type
        for worker in worker_team:
            needed_count = worker.count
            offer_stack = self._timeline[worker.get_agent_id()]
            # Traverse list while not enough resources and grab it
            ind = len(offer_stack) - 1
            while needed_count > 0:
                offer_time, offer_count = offer_stack[ind]
                max_agent_time = max(max_agent_time, offer_time)

                if needed_count < offer_count:
                    offer_count = needed_count
                needed_count -= offer_count
                ind -= 1

        return max(max_agent_time, max_parent_time)

    def update_timeline(self,
                        task_index: int,
                        finish_time: Time,
                        node: GraphNode,
                        node2swork: Dict[GraphNode, ScheduledWork],
                        worker_team: List[Worker]):
        """
        Adds given `worker_team` to the timeline at the moment `finish`
        :param task_index:
        :param finish_time:
        :param node:
        :param node2swork:
        :param worker_team:
        :return:
        """
        # For each worker type consume the nearest available needed worker amount
        # and re-add it to the time when current work should be finished.
        # Addition performed as step in bubble-sort algorithm.
        for worker in worker_team:
            needed_count = worker.count
            worker_timeline = self._timeline[(worker.contractor_id, worker.name)]
            # Consume needed workers
            while needed_count > 0:
                next_time, next_count = worker_timeline.pop()
                if next_count > needed_count:
                    worker_timeline.append((next_time, next_count - needed_count))
                    break
                needed_count -= next_count

            # Add to the right place
            # worker_timeline.append((finish, worker.count))
            # worker_timeline.sort(reverse=True)
            worker_timeline.append((finish_time, worker.count))
            ind = len(worker_timeline) - 1
            while ind > 0 and worker_timeline[ind][0] > worker_timeline[ind - 1][0]:
                worker_timeline[ind], worker_timeline[ind - 1] = worker_timeline[ind - 1], worker_timeline[ind]
                ind -= 1

    def schedule(self,
                 task_index: int,
                 node: GraphNode,
                 id2swork: Dict[GraphNode, ScheduledWork],
                 workers: List[Worker],
                 contractor: Contractor,
                 assigned_start_time: Optional[Time] = None,
                 assigned_time: Optional[Time] = None,
                 work_estimator: Optional[WorkTimeEstimator] = None) -> Time:
        inseparable_chain = node.get_inseparable_chain_with_self()
        st = assigned_start_time if assigned_start_time else self.find_min_start_time(node, workers, id2swork)
        if assigned_time:
            exec_times = {n: (Time(0), assigned_time // len(inseparable_chain))
                          for n in inseparable_chain}
            return self._schedule_with_inseparables(id2swork, workers, contractor, inseparable_chain, st, exec_times,
                                                    work_estimator)
        else:
            return self._schedule_with_inseparables(id2swork, workers, contractor, inseparable_chain, st, {},
                                                    work_estimator)

    def __getitem__(self, item: AgentId):
        return self._timeline[item]

    @staticmethod
    def _schedule_with_inseparables(id2swork: Dict[GraphNode, ScheduledWork],
                                    workers: List[Worker],
                                    contractor: Contractor,
                                    inseparable_chain: List[GraphNode],
                                    start_time: Time,
                                    exec_times: Dict[GraphNode, Tuple[Time, Time]],
                                    work_estimator: Optional[WorkTimeEstimator] = None) -> Time:
        """
        Makes ScheduledWork object from `GraphNode` and worker list, assigned `start_end_time`
        and adds it ti given `id2swork`. Also does the same for all inseparable nodes starts from this one
        :param id2swork:
        :param workers:
        :param contractor:
        :param inseparable_chain:
        :param start_time:
        :param exec_times:
        :param work_estimator:
        :return:
        """
        c_ft = start_time
        for dep_node in inseparable_chain:
            # set start time as finish time of original work
            # set finish time as finish time + working time of current node with identical resources
            # (the same as in original work)
            # set the same workers on it
            # TODO Decide where this should be
            max_parent_time = max((id2swork[pnode].finish_time
                                   for pnode in dep_node.parents),
                                  default=Time(0))
            working_time = exec_times.get(dep_node, None)
            start_time = max(c_ft, max_parent_time)
            if working_time is None:
                working_time = calculate_working_time(dep_node.work_unit, workers, work_estimator)
            new_finish_time = start_time + working_time

            id2swork[dep_node] = ScheduledWork(work_unit=dep_node.work_unit,
                                               start_end_time=(start_time, new_finish_time),
                                               workers=workers,
                                               contractor=contractor)
            # change finish time for using workers
            c_ft = new_finish_time

        return c_ft



from copy import deepcopy
from operator import attrgetter, itemgetter
from typing import Dict, List

from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph
from sampo.schemas.schedule import ScheduledWork, Schedule
from sampo.schemas.time import Time
from sampo.utilities.collections import build_index


def validate_schedule(schedule: Schedule, wg: WorkGraph, contractors: List[Contractor]) -> None:
    """
    Checks if schedule is correct and can be executed.
    If there is an error, this function raises AssertException with an appropriate message
    If it finishes without any exception, it means successful passing of the verification
    :param contractors:
    :param wg:
    :param schedule: to apply verification to
    """
    # checking preconditions
    # check_all_workers_have_same_qualification(schedule.workGraph, schedule.resourcePools)

    # checking the schedule itself
    _check_all_tasks_scheduled(schedule, wg)
    _check_parent_dependencies(schedule, wg)
    _check_all_tasks_have_valid_duration(schedule)
    _check_all_workers_correspond_to_worker_reqs(schedule)
    _check_all_allocated_workers_do_not_exceed_capacity_of_contractors(schedule, contractors)


def _check_all_tasks_scheduled(schedule: Schedule, wg: WorkGraph) -> None:
    # 1. each task of the graph is scheduled
    scheduled_works = {work.work_unit.id: work for work in schedule.works}
    absent_works = [node for node in wg.nodes if node.work_unit.id not in scheduled_works]

    assert len(absent_works) == 0, \
        f"Found works that are not scheduled:\n" \
        f"\t{[work.work_unit.id for work in absent_works]}"


def _check_parent_dependencies(schedule: Schedule, wg: WorkGraph) -> None:
    scheduled_works: Dict[str, ScheduledWork] = {work.work_unit.id: work for work in schedule.works}

    for node in wg.nodes:
        start, end = scheduled_works[node.work_unit.id].start_end_time
        for pnode in node.parents:
            pstart, pend = scheduled_works[pnode.work_unit.id].start_end_time
            assert pstart <= pend <= start <= end


def _check_all_tasks_have_valid_duration(schedule: Schedule) -> None:
    # 3. check if all tasks have duration appropriate for their working hours
    service_works_with_incorrect_duration = [
        work for work in schedule.works
        if work.work_unit.is_service_unit and work.duration != 0
    ]

    assert len(service_works_with_incorrect_duration) == 0, \
        f"Found service works that have non-zero duration:\n" \
        f"\t{[work.work_unit.id for work in service_works_with_incorrect_duration]}"

    # # TODO: make correct duration calculation
    # works_with_incorrect_duration = [
    #     work for work in schedule.works
    #     if not work.work_unit.is_service_unit and work.duration <= 0
    # ]
    #
    # assert len(works_with_incorrect_duration) == 0, \
    #     f"Found works that have incorrect duration:\n" \
    #     f"\t{[work.work_unit.id for work in works_with_incorrect_duration]}"


def _check_all_allocated_workers_do_not_exceed_capacity_of_contractors(schedule: Schedule,
                                                                       contractors: List[Contractor]) -> None:
    # Dict[contractor_id, Dict[worker_name, worker_count]]
    initial_contractors_state: Dict[str, Dict[str, int]] = {}
    for contractor in contractors:
        initial_contractors_state[contractor.id] = {}
        for w in contractor.workers.values():
            if w.name in initial_contractors_state[contractor.id].keys():
                initial_contractors_state[contractor.id][w.name] += w.count
            else:
                initial_contractors_state[contractor.id][w.name] = w.count
    contractors_state = deepcopy(initial_contractors_state)

    check_all_allocated_workers_do_not_exceed_capacity_of_contractors(schedule,
                                                                      initial_contractors_state,
                                                                      contractors_state)


def check_all_allocated_workers_do_not_exceed_capacity_of_contractors(schedule: Schedule,
                                                                      initial_worker_pool: dict[str, dict[str, int]],
                                                                      cur_worker_pool: dict[str, dict[str, int]]) \
        -> dict[str, dict[str, int]]:
    # 4. at each moment sum of allocated workers of all tasks for the same contractor
    # does not exceed capacity of this contractor
    ordered_start_end_events = sorted(
        (el for work in schedule.works
         for el in [("start", work.start_time, work), ("end", work.finish_time, work)]),
        key=itemgetter(1)
    )

    moment_pool: Dict[str, Dict[str, int]] = {}
    moment = Time(0)
    for event_type, time, work in ordered_start_end_events:
        if len(work.workers) == 0:
            continue
        cont = work.workers[0].contractor_id
        cpool = moment_pool.get(cont, {})
        if len(cpool) == 0:
            moment_pool[cont] = cpool

        if time != moment:
            for contractor_id, contractor_pool in moment_pool.items():
                for worker_name, worker_count in contractor_pool.items():
                    available = cur_worker_pool[contractor_id][worker_name]
                    # check
                    assert available + worker_count >= 0, \
                        f"Overuse of workers (event type {event_type} at [{time}]) for contractor '{contractor_id}' " \
                        f"and worker type '{worker_name}': available {available}," \
                        f" while being allocated {worker_count}"
                    assert available + worker_count <= initial_worker_pool[contractor_id][worker_name], \
                        f"Excessive workers appear for contractor '{contractor_id}' and worker type '{worker_name}':" \
                        f"available {available}, being returned {worker_count}, " \
                        f"initial contractor capacity {initial_worker_pool[contractor_id][worker_name]}"

                    # update
                    cur_worker_pool[contractor_id][worker_name] = available + worker_count
            moment_pool.clear()
            moment = time

        cpool = moment_pool.get(cont, {})
        if len(cpool) == 0:
            moment_pool[cont] = cpool

        # grab event to the current equivalency class
        if event_type == "start":
            for w in work.workers:
                cpool[w.name] = cpool.get(w.name, 0) - w.count
        elif event_type == "end":
            for w in work.workers:
                cpool[w.name] = cpool.get(w.name, 0) + w.count
        else:
            raise ValueError(f"Incorrect event type: {event_type}. Only 'start' and 'end' are supported.")

    return cur_worker_pool


def _check_all_workers_correspond_to_worker_reqs(schedule: Schedule):
    for swork in schedule.works:
        worker2req = build_index(swork.work_unit.worker_reqs, attrgetter('kind'))
        for worker in swork.workers:
            req = worker2req[worker.name]
            assert req.min_count <= worker.count <= req.max_count


def _check_all_workers_have_same_qualification(wg: WorkGraph, contractors: List[Contractor]):
    # 1. all workers of the same category belonging to the same contractor should have the same characteristics
    for c in contractors:
        assert all(ws.count >= 1 for _, ws in c.workers.items()), \
            f"There should be only one worker for the same worker category"

    # добавляем агентов в словарь
    agents = {}
    for contractor in contractors:
        for name, val in contractor.workers.items():
            if name[0] not in agents:
                agents[name[0]] = 0
            agents[name[0]] += val.count
    # 2. all tasks should have worker reqs that can be satisfied by at least one contractor
    for v in wg.nodes:
        assert any(
            all(c.worker_types[wreq.kind][0].count
                >= (wreq.min_count + min(agents[wreq.kind], wreq.max_count)) // 2
                for wreq in v.work_unit.worker_reqs)
            for c in contractors
        ), f"The work unit with id {v.work_unit.id} cannot be satisfied by any contractors"



import pandas as pd

from sampo.schemas.time import Time


def fix_split_tasks(baps_schedule_df: pd.DataFrame) -> pd.DataFrame:
    """
    Process and merge information for all tasks, which were separated on the several stages during split
    :param baps_schedule_df: pd.DataFrame: schedule with info for tasks separated on stages
    :return: pd.DataFrame: schedule with merged info for all real tasks
    """
    df_len = baps_schedule_df.shape[0]
    baps_schedule_df.index = range(df_len)

    df = pd.DataFrame(columns=baps_schedule_df.columns)
    unique_ids = set([x.split('_')[0] for x in baps_schedule_df.task_id])

    for task_id in unique_ids:
        task_stages_df = baps_schedule_df.loc[baps_schedule_df.loc[:, 'task_id'].str.contains(task_id)]
        task_series = merge_split_stages(task_stages_df.reset_index(drop=True))
        df.loc[df.shape[0]] = task_series  # append

    df = df.sort_values(by=['start', 'task_name'])
    df['idx'] = range(len(df))

    return df


def merge_split_stages(task_df: pd.DataFrame) -> pd.Series:
    """
    Merge split stages of the same real task into one
    :param task_df: pd.DataFrame: one real task's stages dataframe, sorted by start time
    :return: pd.Series with the full information about the task
    """
    if len(task_df) == 1:
        return task_df.copy().loc[0, :]
    else:
        df = task_df.copy()
        df = df.iloc[-1:].reset_index(drop=True)
        for column in ['task_id', 'task_name']:
            df.loc[0, column] = df.loc[0, column].split('_')[0]  # fix task id and name

        # sum up volumes through all stages
        df.loc[0, 'volume'] = sum(task_df.loc[:, 'volume'])
        df.loc[0, 'workers'] = task_df.loc[0, 'workers']

        # fix connections through all stages
        fixed_connections_lst = []
        for connections_lst in task_df.loc[:, 'successors']:
            for connection in connections_lst:
                if connection[1] != 'IFS':
                    fixed_connections_lst.append(tuple([connection[0].split('_')[0], connection[1]]))
        fixed_connections_lst = list(set(fixed_connections_lst))
        old_successors = df.loc[0, 'successors']
        old_successors_type = type(df.loc[0, 'successors'])
        df.loc[0, 'successors'] = fixed_connections_lst

        # fix task's start time and duration
        df.loc[0, 'start'] = task_df.loc[0, 'start']
        df.loc[0, 'finish'] = task_df.loc[len(task_df) - 1, 'finish']
        df.loc[0, 'duration'] = (df.loc[0, 'finish'] - df.loc[0, 'start']).days + 1

        return df.loc[0, :]


def remove_service_tasks(service_schedule_df: pd.DataFrame) -> pd.DataFrame:
    """
    Remove 'start', 'finish' and milestone tasks from the schedule
    :param service_schedule_df: pd.DataFrame: schedule (with merges stages in the case of baps) with service tasks
    :return: pd.DataFrame: schedule without information about service tasks
    """
    schedule_df = service_schedule_df.copy()

    # Prepare list with service tasks ids
    service_tasks_ids = set(schedule_df.loc[schedule_df.loc[:, 'task_name'].str.contains(
        'start|finish|марке')].loc[:, 'task_id'])

    # Remove rows with service tasks from DataFrame
    schedule_df = schedule_df.loc[~schedule_df.loc[:, 'task_name'].str.contains('start|finish|марке')]

    # Fix connections linked to the service tasks
    fixed_connections_lst = []
    for connections_lst in schedule_df.loc[:, 'successors']:
        fixed_connections_lst.append([])
        for connection in connections_lst:
            if connection[0] not in service_tasks_ids:
                fixed_connections_lst[-1].append(connection)
    schedule_df.loc[:, 'successors'] = fixed_connections_lst
    return schedule_df